/* GNU FaXile - fax image viewer
 *
 * Copyright (C) 2000 Wolfgang Sourdeau
 *
 * Author: Wolfgang Sourdeau <wolfgang@ultim.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <tiffio.h>

#include "tiffimages.h"

inline void setpixelto (char *bitmap, int number, int value)
{
	char *byte;

	byte = bitmap + (int) (number / 8);

	if (value)
		*byte |= (1 << (number % 8));
	else
		*byte &= ~(1 << (number % 8));
}

inline int getpixel (char *bitmap, int width, int factx, int facty,
		     int number)
{
	int division, modulo;
	int x, y, off_x;
	int pixel;
	char *byte;

	pixel = 0;

	if (factx < 1) factx = 1;
	if (facty < 1) facty = 1;

	for (y=0; (y<facty && pixel == 0); y++) {
		off_x = y * width;
		for (x=off_x; (x<(off_x+factx) && pixel == 0); x++) {
			division = (number + x ) / 8;
			modulo = (number + x ) % 8;
			byte = bitmap + division;
			pixel = ((*byte & (1 << modulo)) != 0);
		}
	}

	return pixel;
}

FaxFile *init_file(char* filename)
{
	TIFF *tiff_file;
	FaxFile *fax_file;
	FaxImage *curimg, *newimg;
	int is_fax;
	unsigned int count;
	float xres, yres;
	uint32 width, height;

	fax_file = NULL;

	tiff_file = TIFFOpen (filename, "r");

	if (tiff_file != NULL) {
		TIFFGetField (tiff_file, TIFFTAG_FAXMODE, &is_fax);

		if (is_fax == 1) {
			fax_file = (FaxFile*) g_malloc(sizeof(FaxFile));
			fax_file->tiff_file = tiff_file;
			
			TIFFGetField (tiff_file, TIFFTAG_XRESOLUTION, &xres);
			TIFFGetField (tiff_file, TIFFTAG_YRESOLUTION, &yres);

			fax_file->xres = (unsigned int) xres;
			fax_file->yres = (unsigned int) yres;

			count = 0;
			curimg = g_malloc (sizeof (FaxImage));
			fax_file->first = curimg;
			curimg->prev = NULL;
			TIFFGetField (tiff_file, TIFFTAG_IMAGEWIDTH,
				      &width);
			TIFFGetField (tiff_file, TIFFTAG_IMAGELENGTH,
				      &height);
			curimg->nbr = 0;
			curimg->width = width;
			curimg->height = height;
			curimg->image = NULL;

			while (TIFFReadDirectory (tiff_file)) {
				count++;
				newimg = g_malloc (sizeof (FaxImage));
				TIFFGetField (tiff_file, TIFFTAG_IMAGEWIDTH,
					      &width);
				TIFFGetField (tiff_file, TIFFTAG_IMAGELENGTH,
					      &height);
				newimg->nbr = count;
				newimg->width = width;
				newimg->height = height;
				newimg->image = NULL;
				newimg->prev = curimg;
				curimg->next = newimg;
				curimg = newimg;
			}
			curimg->next = NULL;

			fax_file->nbr_images = count;
		}
	}

	return fax_file;
}

void load_tiff_image (FaxFile *fax_file, FaxImage *fax_image)
{
	tstrip_t strip, nbr_strips;
	tsize_t stripsize;

	TIFFSetDirectory (fax_file->tiff_file, fax_image->nbr);
		
	stripsize = TIFFStripSize(fax_file->tiff_file);
	nbr_strips = TIFFNumberOfStrips(fax_file->tiff_file);
	fax_image->image = _TIFFmalloc(stripsize * nbr_strips);

	for (strip = 0; strip < nbr_strips; strip++) {       
		TIFFReadEncodedStrip(fax_file->tiff_file, strip,
				     fax_image->image + (strip *
							 stripsize),
				     (tsize_t) -1);
	}

	TIFFReverseBits (fax_image->image, (fax_image->width *
					    fax_image->height / 8));
}

void unload_tiff_image (FaxImage *fax_image)
{
	_TIFFfree (fax_image->image);
	fax_image->image = NULL;
}

FaxImage *get_fax_image (FaxFile *fax_file, unsigned int number)
{
	FaxImage *image;
	int count;

	image = fax_file->first;
	for (count = 0; count < number; count++)
		image = image->next;

	return image;
}

void close_fax_file (FaxFile *fax_file)
{
	TIFFClose (fax_file->tiff_file);
	g_free (fax_file);
}

void destroy_fax_image (FaxImage *fax_image)
{
	if (fax_image->image != NULL)
		_TIFFfree (fax_image->image);
	g_free (fax_image);
}

FaxImage *zoomed_fax_image (FaxImage *orig,
			    int new_width, int new_height,
			    int reversed)
{
	int cnt_x, cnt_y, off_x_orig, off_x_dest, pixel;
	int tiff_off;
	float factx, facty;
	char *char_ptr_new, *char_ptr;
	FaxImage *new_image;

	if (new_width == 0) {
		factx = facty = (float) orig->height / (float)
			new_height;
		new_width = orig->width / factx;
	} else if (new_height == 0) {
		facty = factx = (float) orig->width / (float)
			new_width;
		new_height = orig->height / facty;
	} else {
		factx = (float) orig->width / (float) new_width;
		facty = (float) orig->height / (float) new_height;
	}

	new_image = g_malloc (sizeof (FaxImage));
	new_image->nbr = orig->nbr;
	new_image->width = new_width;
	new_image->height = new_height;
	new_image->image = _TIFFmalloc (new_width * new_height);
	new_image->prev = new_image->next = NULL;

	char_ptr_new = new_image->image;
	char_ptr = orig->image;

	if ((new_width % 8) != 0)
		tiff_off = (8 - (new_width % 8));
	else
		tiff_off = 0;

	if (!reversed) {
		for (cnt_y = 0; cnt_y < new_height; cnt_y++) {
			off_x_dest = cnt_y * (new_width + tiff_off);
			off_x_orig = (int) (cnt_y * facty) * orig->width;
			for (cnt_x = 0; cnt_x < new_width; cnt_x++) {
				pixel = getpixel (char_ptr, orig->width,
						  (int) factx, (int) facty,
						  off_x_orig
						  + (int) (cnt_x * factx));
				
				setpixelto (char_ptr_new, off_x_dest + cnt_x,
					    pixel);
			}
		}
	} else {
		for (cnt_y = 0; cnt_y < new_height; cnt_y++) {
			off_x_dest = (new_height - cnt_y) * (new_width + tiff_off);
			off_x_orig = (int) (cnt_y * facty) * orig->width;
			for (cnt_x = 0; cnt_x < new_width; cnt_x++) {
				pixel = getpixel (char_ptr, orig->width,
						  (int) factx, (int) facty,
						  off_x_orig
						  + (int) (cnt_x * factx));
				
				setpixelto (char_ptr_new, off_x_dest +
					    new_width - cnt_x,
					    pixel);
			}
		}
	}

	return new_image;
}

